@@ -1,19 +1,16 @@
 import base64
 import hashlib
 import logging
-import os
 import asyncio
-import random
 import socket
 import struct
-import time
 import typing
 
 import requests
 
-from .sync import persistent_state_ttl, choose_key_block, sync
+from .sync import choose_key_block, sync
 from .utils import init_mainnet_block, init_testnet_block
-from pytoniq_core.boc import Slice, Cell, Builder
+from pytoniq_core.boc import Slice, Cell
 from pytoniq_core.proof.check_proof import check_block_header_proof, check_shard_proof, check_account_proof, check_proof, \
     check_block_signatures, compute_validator_set
 from pytoniq_core.boc.address import Address
@@ -22,7 +19,7 @@ from pytoniq_core.crypto.ciphers import Server, Client, get_random, create_aes_c
 from pytoniq_core.crypto.crc import crc16
 
 from pytoniq_core.tl.generator import TlGenerator, TlSchema
-from pytoniq_core.tl.block import BlockId, BlockIdExt
+from pytoniq_core.tl.block import BlockIdExt, BlockId  # do not remove this import!
 from pytoniq_core.tlb.config import ConfigParam34, ConfigParam28, ConfigParam
 from pytoniq_core.tlb.transaction import Transaction
 from pytoniq_core.tlb.utils import deserialize_shard_hashes
@@ -87,12 +84,8 @@ class LiteClient:
         self.delta = 0.02  # listen delay
 
         self.listener: asyncio.Task = None
-        rand_int = str(random.randint(10**9, 3 * 10**9))
-        self.listener_name: str = 'listener' + rand_int
         self.pinger: asyncio.Task = None
-        self.pinger_name: str = 'pinger' + rand_int
         self.updater: asyncio.Task = None
-        self.updater_name: str = 'updater' + rand_int
 
         """########### TL ###########"""
         if tl_schemas_path is None:
@@ -155,20 +148,23 @@ class LiteClient:
 
             qid = result.get('query_id', result.get('random_id'))  # return query_id for ordinary requests, random_id for ping-pong requests, None for handshake
 
-            request = self.tasks.pop(qid)
+            request: asyncio.Future = self.tasks.pop(qid)
 
             result = result.get('answer', {})
-            request.set_result(result)
+            if not request.done():
+                request.set_result(result)
 
     async def connect(self) -> None:
+        if self.inited:
+            raise LiteClientError('The client is already connected')
         self.loop = asyncio.get_running_loop()
         handshake = self.handshake()
         self.reader, self.writer = await asyncio.wait_for(asyncio.open_connection(self.server.host, self.server.port), self.timeout)
         future = await asyncio.wait_for(self.send(handshake, None), self.timeout)
-        self.listener = asyncio.create_task(self.listen(), name=self.listener_name)
+        self.listener = asyncio.create_task(self.listen())
         await self.update_last_blocks()
-        self.pinger = asyncio.create_task(self.ping(), name=self.pinger_name)
-        self.updater = asyncio.create_task(self.block_updater(), name=self.updater_name)
+        self.pinger = asyncio.create_task(self.ping())
+        self.updater = asyncio.create_task(self.block_updater())
         await future
         self.inited = True
 
@@ -177,9 +173,16 @@ class LiteClient:
         await self.connect()
 
     async def close(self) -> None:
-        for i in asyncio.all_tasks(self.loop):
-            if i.get_name() in {self.listener_name, self.pinger_name, self.updater_name}:
-                i.cancel()
+        for i in [self.pinger, self.updater, self.listener]:
+            i.cancel()
+            while not i.done():
+                await asyncio.sleep(0.001)
+        self.inited = False
+        self.tasks = {}
+        self.reader = None
+        self.writer.close()
+        await self.writer.wait_closed()
+        self.writer = None
         self.logger.info(msg='client has been closed')
 
     def handshake(self) -> bytes:
@@ -287,7 +290,10 @@ class LiteClient:
         if self.last_mc_block is None:
             self.last_mc_block = await self.get_trusted_last_mc_block()
         while True:
-            await self.wait_masterchain_seqno(self.last_mc_block.seqno + 1, timeout_ms=10000, schema_name='getMasterchainInfo', data={})
+            try:
+                await self.wait_masterchain_seqno(self.last_mc_block.seqno + 1, timeout_ms=10000, schema_name='getMasterchainInfo', data={})
+            except asyncio.TimeoutError:
+                continue
             await self.update_last_blocks()
 
     async def get_masterchain_info(self):
@@ -324,12 +330,14 @@ class LiteClient:
     async def get_version(self):
         return await self.liteserver_request('getVersion', {})
 
-    async def get_state(self, wc: int, shard: typing.Optional[int], seqno: int, root_hash: typing.Union[str, bytes], file_hash: typing.Union[str, bytes]):
-        # TODO doesnt work: {'code': -400, 'message': 'cannot request total state: possibly too large'}
+    async def get_state(self, wc: int, shard: typing.Optional[int],
+                        seqno: int, root_hash: typing.Union[str, bytes],
+                        file_hash: typing.Union[str, bytes]
+                        ) -> dict:
         block = self.pack_block_id_ext(wc=wc, shard=shard, seqno=seqno, root_hash=root_hash, file_hash=file_hash)
         return await self.liteserver_request('getState', block)
 
-    async def raw_get_block_header(self, block: BlockIdExt):
+    async def raw_get_block_header(self, block: BlockIdExt) -> Block:
         result = await self.liteserver_request('getBlockHeader', {'id': block.to_dict()} | {'mode': 0})
         h_proof = Cell.one_from_boc(result['header_proof'])
         block_id = BlockIdExt.from_dict(result['id'])
@@ -342,12 +350,24 @@ class LiteClient:
                 await self.get_mc_block_proof(known_block=self.last_key_block, target_block=block_id)
         return Block.deserialize(h_proof[0].begin_parse())
 
-    async def get_block_header(self, wc: int, shard: typing.Optional[int], seqno: int, root_hash: typing.Union[str, bytes], file_hash: typing.Union[str, bytes]):
+    async def get_block_header(self, wc: int, shard: typing.Optional[int], seqno: int,
+                               root_hash: typing.Union[str, bytes],
+                               file_hash: typing.Union[str, bytes]
+                               ) -> Block:
+        """
+        :param wc: block workchain
+        :param shard: block shard
+        :param seqno: block seqno
+        :param root_hash: block root hash
+        :param file_hash: block file hash
+        :return: block header
+        """
         block = self.pack_block_id_ext(wc=wc, shard=shard, seqno=seqno, root_hash=root_hash, file_hash=file_hash)
         return await self.raw_get_block_header(BlockIdExt.from_dict(block))
 
     async def lookup_block(self, wc: int, shard: int, seqno: int = -1,
-                           lt: typing.Optional[int] = None, utime: typing.Optional[int] = None) -> typing.Tuple[BlockIdExt, Block]:
+                           lt: typing.Optional[int] = None,
+                           utime: typing.Optional[int] = None) -> typing.Tuple[BlockIdExt, Block]:
         """
         :param wc: block workchain
         :param shard: block shard
@@ -390,11 +410,15 @@ class LiteClient:
                 await self.prove_block(block_id)
         return Block.deserialize(result_block.begin_parse())
 
-    async def get_block(self, wc: int, shard: typing.Optional[int], seqno: int, root_hash: typing.Union[str, bytes], file_hash: typing.Union[str, bytes]):
+    async def get_block(self, wc: int, shard: typing.Optional[int],
+                        seqno: int, root_hash: typing.Union[str, bytes],
+                        file_hash: typing.Union[str, bytes]) -> Block:
         block = self.pack_block_id_ext(wc=wc, shard=shard, seqno=seqno, root_hash=root_hash, file_hash=file_hash)
         return await self.raw_get_block(BlockIdExt.from_dict(block))
 
-    async def raw_get_account_state(self, address: typing.Union[str, Address], block: typing.Optional[BlockIdExt] = None) -> typing.Tuple[typing.Optional[Account], typing.Optional[ShardAccount]]:
+    async def raw_get_account_state(self, address: typing.Union[str, Address],
+                                    block: typing.Optional[BlockIdExt] = None
+                                    ) -> typing.Tuple[typing.Optional[Account], typing.Optional[ShardAccount]]:
         trusted = False
         if block is None or block == self.last_mc_block:
             block = self.last_mc_block
@@ -428,9 +452,13 @@ class LiteClient:
             address = Address(address)
         return SimpleAccount.from_raw((await self.raw_get_account_state(address))[0], address)
 
-    async def run_get_method(self, address: typing.Union[Address, str], method: typing.Union[int, str], stack: list) -> list:
+    async def run_get_method(self, address: typing.Union[Address, str],
+                             method: typing.Union[int, str], stack: list,
+                             block: BlockIdExt = None
+                             ) -> list:
         mode = 7  # 111
-        block = self.last_mc_block
+        if block is None:
+            block = self.last_mc_block
 
         if isinstance(address, str):
             address = Address(address)
@@ -459,7 +487,10 @@ class LiteClient:
 
         return VmStack.deserialize(Slice.one_from_boc(result['result']))
 
-    async def raw_get_shard_info(self, block: typing.Optional[BlockIdExt] = None, wc: int = 0, shard: int = -9223372036854775808, exact: bool = True):
+    async def raw_get_shard_info(self, block: typing.Optional[BlockIdExt] = None,
+                                 wc: int = 0, shard: int = -9223372036854775808,
+                                 exact: bool = True
+                                 ) -> ShardDescr:
         trusted = False
         if block is None or block == self.last_mc_block:
             block = self.last_mc_block
@@ -538,7 +569,9 @@ class LiteClient:
                 )
         return result
 
-    async def get_one_transaction(self, address: typing.Union[Address, str], lt: int, block: BlockIdExt) -> typing.Optional[Transaction]:
+    async def get_one_transaction(self, address: typing.Union[Address, str],
+                                  lt: int, block: BlockIdExt
+                                  ) -> typing.Optional[Transaction]:
         if isinstance(address, str):
             address = Address(address)
 
@@ -568,7 +601,9 @@ class LiteClient:
 
         return Transaction.deserialize(transaction_root.begin_parse())
 
-    async def raw_get_transactions(self, address: typing.Union[Address, str], count: int, from_lt: int = None, from_hash: typing.Optional[bytes] = None) -> typing.Tuple[typing.List[Transaction], typing.List[BlockIdExt]]:
+    async def raw_get_transactions(self, address: typing.Union[Address, str], count: int,
+                                   from_lt: int = None, from_hash: typing.Optional[bytes] = None
+                                   ) -> typing.Tuple[typing.List[Transaction], typing.List[BlockIdExt]]:
         if isinstance(address, str):
             address = Address(address)
 
@@ -603,7 +638,17 @@ class LiteClient:
         # assert len(tr_result) == count, f'expected {count} transactions, got {len(tr_result)}'
         return tr_result, block_ids
 
-    async def get_transactions(self, address: typing.Union[Address, str], count: int, from_lt: int = None, from_hash: typing.Optional[bytes] = None) -> typing.List[Transaction]:
+    async def get_transactions(self, address: typing.Union[Address, str], count: int,
+                               from_lt: int = None, from_hash: typing.Optional[bytes] = None
+                               ) -> typing.List[Transaction]:
+        """
+        Returns account transactions
+        :param address:
+        :param count:
+        :param from_lt:
+        :param from_hash:
+        :return:
+        """
         result: typing.List[Transaction] = []
 
         for i in range(0, count, 16):
@@ -611,64 +656,101 @@ class LiteClient:
             tr_result, block_ids = await self.raw_get_transactions(address, amount, from_lt, from_hash)
             result += tr_result
             from_lt, from_hash = result[-1].prev_trans_lt, result[-1].prev_trans_hash
+            if from_lt == 0:
+                break
         # assert len(result) == count, f'expected {count} transactions, got {len(result)}'
         return result
 
-    async def raw_get_block_transactions(self, block: BlockIdExt, count: int = 256) -> typing.List[dict]:
+    async def raw_get_block_transactions(self, block: BlockIdExt, count: int = 1024) -> typing.List[dict]:
+
+        def parse_transactions(result: dict):
+            if not result['ids']:
+                return []
+
+            transactions_ids = result['ids']
+            if self.trust_level <= 1:
+                proof = Cell.one_from_boc(result['proof'])
+                check_block_header_proof(proof[0], block.root_hash)
+                acc_block = Block.deserialize(proof[0].begin_parse()).extra.account_blocks[0]
+
+            for tr in transactions_ids:
+                tr['hash'] = bytes.fromhex(tr['hash'])
+                tr.pop('mode')  # in this lib mode is a fixed num, so we don't really need it in result, moreover mode can mislead
+                if self.trust_level <= 1:
+                    block_trs: dict = acc_block.get(int(tr['account'], 16)).transactions[0]
+                    block_tr: Cell = block_trs.get(tr['lt'])
+                    assert block_tr.get_hash(0) == tr['hash']
+                tr['account'] = Address((block.workchain, bytes.fromhex(tr['account'])))
+
+            return transactions_ids
+
         mode = 39  # 100111
         data = {'id': block.to_dict(), 'mode': mode, 'count': count, 'want_proof': b''}
         result = await self.liteserver_request('listBlockTransactions', data)
-        if not result['ids']:
-            return []
-        transactions_ids = result['ids']
-        if self.trust_level <= 1:
-            proof = Cell.one_from_boc(result['proof'])
-            check_block_header_proof(proof[0], block.root_hash)
-            acc_block = Block.deserialize(proof[0].begin_parse()).extra.account_blocks[0]
 
         if not self.trust_level and block != self.last_mc_block:
             await self.prove_block(block)
 
-        for tr in transactions_ids:
-            tr['hash'] = bytes.fromhex(tr['hash'])
-            tr.pop('mode')  # in this lib mode is a fixed num, so we don't really need it in result, moreover mode can mislead
+        transactions = parse_transactions(result)
+
+        while result['incomplete']:
+            mode = 167  # 10100111
+            data |= {'mode': mode, 'after': {'account': transactions[-1]['account'].hash_part.hex(), 'lt': transactions[-1]['lt']}}
+            result = await self.liteserver_request('listBlockTransactions', data)
+            transactions += parse_transactions(result)
+
+        return transactions
+
+    async def raw_get_block_transactions_ext(self, block: BlockIdExt, count: int = 1024) -> typing.List[Transaction]:
+
+        def parse_transactions(result: dict):
+            if not result['transactions']:
+                return []
+
+            transactions_cells = Cell.from_boc(result['transactions'])
+
             if self.trust_level <= 1:
-                block_trs: dict = acc_block.get(int(tr['account'], 16)).transactions[0]
-                block_tr: Cell = block_trs.get(tr['lt'])
-                assert block_tr.get_hash(0) == tr['hash']
-            tr['account'] = Address((block.workchain, bytes.fromhex(tr['account'])))
+                proof = Cell.one_from_boc(result['proof'])
+                check_block_header_proof(proof[0], block.root_hash)
+                acc_block = Block.deserialize(proof[0].begin_parse()).extra.account_blocks[0]
+            tr_result = []
 
-        return transactions_ids
+            for tr_root in transactions_cells:
+                transaction = Transaction.deserialize(tr_root.begin_parse())
+                if self.trust_level <= 1:
+                    prunned_tr_cell = acc_block.get(int(transaction.account_addr_hex, 16)).transactions[0].get(
+                        transaction.lt)
+                    assert prunned_tr_cell.get_hash(0) == tr_root.get_hash(0)
+                # transaction.account = Address((block.workchain, bytes.fromhex(transaction.account_addr)))
+                tr_result.append(transaction)
+
+            return tr_result
 
-    async def raw_get_block_transactions_ext(self, block: BlockIdExt, count: int = 256) -> typing.List[Transaction]:
         mode = 39  # 100111
         data = {'id': block.to_dict(), 'mode': mode, 'count': count, 'want_proof': b''}
         result = await self.liteserver_request('listBlockTransactionsExt', data)
 
-        if not result['transactions']:
-            return []
+        if not self.trust_level and block != self.last_mc_block:
+            await self.prove_block(block)
 
-        transactions_cells = Cell.from_boc(result['transactions'])
+        transactions = parse_transactions(result)
 
-        if self.trust_level <= 1:
-            proof = Cell.one_from_boc(result['proof'])
-            check_block_header_proof(proof[0], block.root_hash)
-            acc_block = Block.deserialize(proof[0].begin_parse()).extra.account_blocks[0]
-            if not self.trust_level and block != self.last_mc_block:
-                await self.prove_block(block)
-        tr_result = []
+        while result['incomplete']:
+            mode = 167  # 10100111
+            data |= {'mode': mode, 'after': {'account': transactions[-1].account_addr_hex, 'lt': transactions[-1].lt}}
+            result = await self.liteserver_request('listBlockTransactionsExt', data)
+            transactions += parse_transactions(result)
 
-        for tr_root in transactions_cells:
-            transaction = Transaction.deserialize(tr_root.begin_parse())
-            if self.trust_level <= 1:
-                prunned_tr_cell = acc_block.get(int(transaction.account_addr_hex, 16)).transactions[0].get(transaction.lt)
-                assert prunned_tr_cell.get_hash(0) == tr_root.get_hash(0)
-            # transaction.account = Address((block.workchain, bytes.fromhex(transaction.account_addr)))
-            tr_result.append(transaction)
+        return transactions
 
-        return tr_result
-
-    async def raw_get_mc_block_proof(self, known_block: BlockIdExt, target_block: typing.Optional[BlockIdExt] = None, return_best_key_block=False) -> typing.Tuple[bool, BlockIdExt, typing.Optional[BlockIdExt], typing.Optional[int]]:
+    async def raw_get_mc_block_proof(self, known_block: BlockIdExt, target_block: typing.Optional[BlockIdExt] = None,
+                                     return_best_key_block=False
+                                     ) -> typing.Tuple[
+                                                        bool,
+                                                        BlockIdExt,
+                                                        typing.Optional[BlockIdExt],
+                                                        typing.Optional[int]
+                                        ]:
         """
         :param known_block: block you trust
         :param target_block: block you want to prove
@@ -749,7 +831,10 @@ class LiteClient:
                     last_trusted = to_block
         return last_trusted == target_block, last_trusted, best_key, best_key_ts
 
-    async def get_mc_block_proof(self, known_block: BlockIdExt, target_block: BlockIdExt, return_best_key_block=False):
+    async def get_mc_block_proof(self, known_block: BlockIdExt,
+                                 target_block: BlockIdExt,
+                                 return_best_key_block=False
+                                 ) -> typing.Tuple[typing.Optional[BlockIdExt], int]:
         self.logger.debug(msg=f'PROOF BLOCKS\nfrom: {known_block}\ntarget: {target_block}')
         last_proved = known_block
         best_key = None
@@ -762,13 +847,13 @@ class LiteClient:
         if return_best_key_block:
             return best_key, best_key_ts
 
-    async def prove_block(self, target_block: BlockIdExt):
+    async def prove_block(self, target_block: BlockIdExt) -> None:
         if target_block.workchain == -1:
             await self.get_mc_block_proof(self.last_key_block, target_block)
         else:
             await self.get_shard_block_proof(target_block, True)
 
-    def unpack_config(self, block: BlockIdExt, config_proof: Cell, state_proof: Cell):
+    def unpack_config(self, block: BlockIdExt, config_proof: Cell, state_proof: Cell) -> dict:
         if self.trust_level <= 1:
             state_hash = check_block_header_proof(state_proof[0], block.root_hash, True)
             if config_proof[0].get_hash(0) != state_hash:
@@ -784,7 +869,7 @@ class LiteClient:
 
         return config_res
 
-    async def get_config_all(self, blk: typing.Optional[BlockIdExt] = None):
+    async def get_config_all(self, blk: typing.Optional[BlockIdExt] = None) -> dict:
         trusted = False
         if blk is None:
             blk = self.last_mc_block
@@ -803,7 +888,7 @@ class LiteClient:
         state_proof = Cell.one_from_boc(result['state_proof'])
         return self.unpack_config(blk, config_proof, state_proof)
 
-    async def get_config_params(self, params: typing.List[int], blk: typing.Optional[BlockIdExt] = None):
+    async def get_config_params(self, params: typing.List[int], blk: typing.Optional[BlockIdExt] = None) -> dict:
         trusted = False
         if blk is None:
             blk = self.last_mc_block
@@ -865,11 +950,10 @@ class LiteClient:
             check_block_header_proof(proof[0], last_shard_blk.root_hash)
             shrd_blk = Block.deserialize(proof[0].begin_parse())
             prev_blk = shrd_blk.info.prev_ref.prev
-
             last_shard_blk = BlockIdExt.from_dict(prev_blk.__dict__ | {'workchain': last_shard_blk.workchain, 'shard': last_shard_blk.shard})
-            if last_shard_blk == blk:
-                return
-            raise LiteClientError('incorrect proof')
+        if last_shard_blk == blk:
+            return
+        raise LiteClientError('incorrect proof')
 
     async def raw_send_message(self, message: bytes):
         data = {'body': message}
@@ -880,7 +964,7 @@ class LiteClient:
     @classmethod
     def from_config(cls, config: dict, ls_i: int = 0, trust_level: int = 2, timeout: int = 10):
         ls = config['liteservers'][ls_i]
-        init_block = config['validator']['init_block']
+        init_block = config['validator']['init_block'].copy()
         init_block['file_hash'] = base64.b64decode(init_block['file_hash']).hex()
         init_block['root_hash'] = base64.b64decode(init_block['root_hash']).hex()
         init_block = BlockIdExt.from_dict(init_block)
